/*****************************************************************************************************/
//
//	   						Compresseur selon la mthode de Huffman
//										par MB
//
//
//	cre en janvier 2000
//
/*****************************************************************************************************/





/*********************   INCLUDES   *************************/

#include <conio.h>
#include <string.h>
#include <stdio.h>
#include <alloc.h>
#include <io.h>
#include <stdlib.h>
#include <dos.h>
#include <fcntl.h>
#include <sys/stat.h>


/********************  DEFINES  ***************************/

#define BUFFER_LENGTH	1024		// tampon pour la lecture ou l'ecriture de caractere

									// code de renvoi des fonctions
#define OK				0
#define ARG_ERROR		1
#define SOURCE_ERROR	2
#define BAD_MEM			3
#define DEST_ERROR		4



/******************** DECLARATION DE TYPES   ***************/


struct S_Tree {
				long poids;			// poids du caractere
				unsigned int car;	// code ASCII du caractere
				S_Tree *Rptr;		// pointeur sur le noeud de droite
				S_Tree *Lptr;		// pointeur sur le noeud de gauche
			};



struct S_File {
				unsigned char nom[13];	// nom du fichier
				unsigned long length;	// taille
				int handle;				// handle associe
			};



struct S_Header {
				unsigned char sign;					// tag
				unsigned char nom_original[13];		// nom du fichier compresse
				unsigned long taille_originale;		// taille originale du fichier compresse
				unsigned int taille_table;			// nombre de caracteres dans la table
				};

											// pour la position du curseur graphique
struct S_Curpos {
				unsigned int x;
				unsigned int y;
				};


/*******************  VARIABLES GLOBALES  *******************/


extern unsigned _stklen = 20000U;			// pile de 20kb

											// liste des messages d'erreur
unsigned char error_message[5] [62] = {
									"Compression termine avec succs !",
									"Erreur dans les arguments.",
									"Erreur d'ouverture ou de lecture du fichier source.",
									"Pas assez de mmoire disponible.",
									"Erreur de cration ou d'criture du fichier de destination."
									};

S_File			source, destination;	// fichiers source et destination
S_Curpos		curpos;					// position du curseur graphique
int				dim_table = 256;		// dimension originale de la table de caracteres
unsigned char	*table_code[256];		// table des codes


/*******************  DECLARATION DE FONCTIONS  ***************/


int	Read_arguments(int nb_arg, char** argv);
int	Compress(void);
int	Creat_tree(S_Tree* table[]);
int	Write_header(unsigned int nb_car);
int	Write_file(void);
int	Write_table(void);
void Help(void);
void Ordonne_table(S_Tree* table[]);
void Creat_code(S_Tree* noeud);
unsigned char Ascii2hex(unsigned char code[], unsigned char offset=0);




/*******************************************************/
/*****************    CODE    **************************/
/*******************************************************/




/*******************************************************/
// main
// role	  : fonction principale
// entree : nombre d'arguments saisis, tableau des arguments
// sortie : rien
//
// gere d'abord les arguments d'entree et appelle la fonction principale de compression
// en tenant compte des erreurs et exceptions
/*******************************************************/
void main(int nb_arg, char** argv)
{

	int status = OK;		// status des appels de fonctions pour les erreurs

	_setcursortype(_NOCURSOR);	// n'affiche pas le curseur graphique
	printf("\n\n\n\t\tCompresseur selon la mthode de Huffman        par MB\n\n\n");


							// lit les arguments
	if( (status = Read_arguments(nb_arg, argv) ) != OK)
	{
		printf("\n  ERREUR : %s\n", error_message[status]);
		printf("  Opration annule !\n\n\n\n");
		Help();

		_setcursortype(_NORMALCURSOR);	// raffiche le curseur

		return;

	}

						// compresse le fichier
	if( (status = Compress() ) != OK)
	{
		printf("\n  ERREUR : %s\n", error_message[status]);
		printf("  Opration annule !\n\n\n\n");

		_setcursortype(_NORMALCURSOR);

		return;
	}


	printf("\n%s\n", error_message[status]);	// normalement dit que tout va bien :)
	_setcursortype(_NORMALCURSOR);


	return;
}



/*******************************************************/
// int Read_arguments
// role    : lit les arguments passes au programme
// entrees : nombre d'arguments, pointeur sur le tableau contenant les arguments
// sorties : OK si les arguments sont corrects, ARG_ERROR s'il n'y en a pas assez ou s'ils sont incorrects
//
// lit les arguments et s'ils sont valides -> affiche l'aide (si demand)
//										   -> capture le nom du fichier source et cree celui de destination
/*******************************************************/
int Read_arguments(int nb_arg, char **argv)
{

	unsigned char *ptr;
	unsigned char ext[]=".huf";
	unsigned char point='.';

								// teste le nombre d'argument
	if (nb_arg<2)				// il doit y en avoir au moins 2
	{
		return ARG_ERROR;
	}
  
								// teste si demande d'aide
	if (!strcmp(argv[1], "/?") )
	{
		Help();
		return OK;
	}

								
	strcpy(source.nom, argv[1]);	// copie le nom du fichier source

									// remplace l'extention .*** par .huf
	if( (ptr = strchr(source.nom, point)) == NULL)
		return ARG_ERROR;

	memset(destination.nom, '\0', 12);							
	strncpy(destination.nom, source.nom,(unsigned int) (ptr-source.nom));
	strcat(destination.nom, ext);

	return OK;
}

/*******************************************************/
// void Help
// role   : affiche l'aide
// entree : rien
// sortie : rien
/*******************************************************/
void Help(void)
{

	printf(" usage : Compress.exe  paramtre\n");
	printf(" param : *   nom du fichier  compresser\n");
	printf("         *   /?\n");

	return;
}




/*******************************************************/
// int Compress
// role   : regroupe toutes les etapes de la compression
// entree : rien
// sortie : OK si pas de problemes, SOURCE_ERROR si erreur de lecture, DEST_ERROR si erreur d'ecriture, BAD_MEM si pas assez de memoire
/*******************************************************/
int Compress(void)
{

	S_Tree *table_occ[256];					// table d'occurence
	int status = OK;						// status des appels de fonction
	int cpt = 0;							// compteur
	unsigned int nb_car = 256;				// nombre de caracteres differents
	unsigned long lcpt = 0L;				// compteur long
	unsigned char scar[BUFFER_LENGTH +1];	// tampon pour la lecture des caracteres




				/* ********   OUVRE LE FICHIER SOURCE  ************ */

	if( (source.handle =(int) open(source.nom, O_RDONLY | O_BINARY, S_IREAD)) == -1)
	{
		return SOURCE_ERROR;
	}
	source.length =(long) filelength(source.handle);		// lit sa longueur




		/* ********   CREE LA TABLE D'OCCURENCE ET L'INITIALISE   ********* */

	for(cpt=0; cpt<256; cpt++)
	{										// alloue de la memoire
		if( (table_occ[cpt] =(S_Tree*) farmalloc(sizeof(S_Tree)) ) == NULL)	
		{
			while(cpt--)					// libere la memoire juste allouee
				farfree(table_occ[cpt]);
			return BAD_MEM;
		}									// initialise la structure
		table_occ[cpt]->poids = 0L;	
		table_occ[cpt]->car =(unsigned int) cpt;
		table_occ[cpt]->Lptr = NULL;
		table_occ[cpt]->Rptr = NULL;
	}



			/* ********   REMPLIE LA TABLE D'OCCURENCE    ******** */

	printf("Lecture du fichier %s : ...", source.nom);
	curpos.x = wherex();
	curpos.y = wherey();

	for(lcpt=0L; lcpt<(long)source.length; lcpt+=(unsigned long)BUFFER_LENGTH)
	{
		if ( (cpt=read(source.handle, scar, BUFFER_LENGTH)) == -1 )		// lit 512 caracteres a la fois
		{
			for(cpt=0; cpt<256; cpt++)		// libere la memoire
				farfree(table_occ[cpt]);
			close(source.handle);
			return SOURCE_ERROR;
		}

		for(cpt--; cpt>=0; cpt--)		// augmente l'occurence dans la table de tous les caracteres lus
			table_occ[scar[cpt]]->poids ++;

		gotoxy(curpos.x, curpos.y);					// affiche l'avancement
		printf(" %u\%", 100*lcpt/source.length);
	}
	gotoxy(curpos.x, curpos.y);
	printf(" Termine.\n");


			/* ******  REDUIT LA TABLE AUX CARACTERES UTILES *********** */

	Ordonne_table(table_occ);
	while(table_occ[dim_table-1]->poids == 0UL)
	{										// libere la mmoire des cellules vides
		dim_table--;
		farfree(table_occ[dim_table]);
	}
					// recupere le nombre de caractere differents
	nb_car = dim_table;		



		/* ***********   CREE L'ARBRE DE HUFFMAN   *********** */

	printf("Creation de l'abre de Huffman ...");
	if( (status=Creat_tree(table_occ)) != OK)
	{
		for(cpt=0; cpt<nb_car; cpt++)
			farfree(table_occ[cpt]);
		close(source.handle);
		return BAD_MEM;
	}
	printf("  Termine.\n");


		/* **********   INITIALISE LA TABLE DES CODES   ********** */	

	for(cpt=0; cpt<256; cpt++)
	{
		table_code[cpt] = (unsigned char *) farmalloc(2);
		strcpy(table_code[cpt], "2");
	}


		/* *********   CREE LES CODES DE CHAQUE CARACTERE   ********** */

	printf("Creation des codes ...");
	Creat_code(table_occ[0]);
	printf("  Termine.\n");


		/* ********   DETRUIT LA TABLE D'OCCURENCE     ********** */

	for(cpt=0; cpt<dim_table; cpt++)
		farfree(table_occ[cpt]);



		/* ********   ECRIT LE HEADER    *********** */

	printf("Ecriture du fichier : %s ...", destination.nom);
	curpos.x = wherex();
	curpos.y = wherey();
	if( (status=Write_header(nb_car)) != OK)
	{
		for(cpt=0; cpt<256; cpt++)
			farfree(table_code[cpt]);

		close(source.handle);
		return status;
	}


		/* ********   ECRIT LA TABLE DES CODES   ******* */

	if( (status=Write_table()) != OK)
	{
		for(cpt=0; cpt<256; cpt++)
			farfree(table_code[cpt]);

		close(source.handle);
		close(destination.handle);
		return status;
	}


		/* ********   ECRIT LE FICHIER EN LE COMPRESSANT   ********** */

	if( (status=Write_file()) != OK)
	{
		for(cpt=0; cpt<256; cpt++)
			farfree(table_code[cpt]);

		close(source.handle);
		close(destination.handle);
		return status;
	}
	gotoxy(curpos.x, curpos.y);
	printf("  Termine.\n");



		/* *******   LIBERE LA MEMOIRE DE LA TABLE DES CODES   ******** */

	printf("Libration de la mmoire et fermeture des fichiers ...");
	for(cpt=0; cpt<256; cpt++)
		farfree(table_code[cpt]);

	close(source.handle);
	close(destination.handle);
	printf("  Termine.\n");

	return OK;
}




/*******************************************************/
// void Ordonne
// role   : tri les cellules de la table suivant les poids dcroissants
// entree : table  trier, dimention de la table (variable  globale)
// sortie : rien
//
// prend comme reference le poids de la premiere cellule et balaye la table pour trouver
// un poids plus fort, si oui, permutation des cellules
/*******************************************************/
void Ordonne_table(S_Tree *table[])
{

	S_Tree *ptree = NULL;			// pointeur temporaire
	int i = 0;						// compteur
	int j = 0;						// compteur
	int flag = 0;
	int offset = 0;	
	unsigned long poids_ref = 0L;


					// balaye toute la table
	for(i=0; i<dim_table; i++)
	{
		poids_ref =(unsigned long) table[i]->poids;		// poids de reference

		for(j=i+1,flag=0; j<dim_table; j++)		// verifie si une cellule plus lourde existe plus loin
		{
			if(table[j]->poids > poids_ref)		// si oui, on la repere et on la prend comme nouvelle reference
			{
				poids_ref = table[j]->poids;
				ptree =(S_Tree*) table[j];			// sauvegarde l'adresse de la cellule
				flag = 1;							// drapeau
				offset = j;							// numero de la cellule
			}
		}

		if(flag)			// permutation des cellules
		{											// echange des adresses
			table[offset] =(S_Tree*) table[i];
			table[i] =(S_Tree*) ptree;
		}
	}
	ptree = NULL;
}




/*******************************************************/
// int Creat_tree
// role   : cree l'arbre de Huffman a partir de la table d'occurence
// entree : table d'occurence
// sortie : OK si pas de probleme, BAD_MEM si manque de memoire
//
// regroupe les 2 derniere cellules en une seule et la replace dans la table comme une cellule normale
// puis recommence jusqu' ce qu'il n'en reste plus qu'une, la racine de l'arbre
/*******************************************************/
int Creat_tree(S_Tree* table[])
{
	S_Tree *ptree = NULL;		// pointeur
	int cpt = dim_table-1;		// compteur


								// on balaye toute la table en commencant par le bas
	while(cpt>0)
	{
		if( (ptree =(S_Tree*) farmalloc(sizeof(S_Tree))) == NULL)	// creation d'une cellule vide
		{
			return BAD_MEM;
		}
		
		ptree->car = 256;						// cette cellule est un noeud
		ptree->Rptr =(S_Tree*) table[cpt--];	// pointeur de droite pointe sur derniere cellule
		ptree->Lptr =(S_Tree*) table[cpt];		// pointeur de gauche pointe sur avant-derniere cellule
		ptree->poids =(unsigned long) ptree->Lptr->poids + ptree->Rptr->poids;	// poids global
		table[cpt] =(S_Tree*) ptree;			// avant dereniere cellule devient le nouveau noeud
		dim_table--;							// la table perd une cellule  ( -2+1 )

		Ordonne_table(table);					// reordonne la table selon les poids
	}
	return OK;
}


/*******************************************************/
// void Creat_code
// role   : cree les code a partir de l'arbre de Huffman
// entree : racine de l'arbre
// sortie : rien
//
// fonction recursive, on examine toute la partie droite du noeud en ajoutant 0 au code  chaque descente
// jusqu' ce que le noeud contienne un caractere, puis on remonte pour examiner toute la partie gauche
/*******************************************************/
void Creat_code(S_Tree* noeud)
{
	static unsigned char code[256];		// code static
	static unsigned char length = 0;	// longueur du code

	length++;							// le code augmente d'un bit

	if(noeud->car == 256)		// si le noeud ne contient pas de caractere
	{
		code[length-1] = '0';		// rajoute 0 au code, 
		Creat_code(noeud->Rptr);	// on explore la partie droite du noeud

		code[length-1] = '1';		// on remplace le 0 par 1
		Creat_code(noeud->Lptr);	// on explore la partie gauche

		farfree(noeud);				// on libere le noeud que l'on vient d'explorer
	}

	else						// si c'est un caractere
	{								// on alloue suffisemment de place dans la table pour stocker le code
		table_code[noeud->car] =(unsigned char*) farrealloc(table_code[noeud->car], strlen(code)+1);
		strcpy(table_code[noeud->car], code);	// on sauvegarde le code en ascii
	}
	code[--length] = '\0';			// efface le dernier caractere et diminue le code d'1 bit
}

/*******************************************************/
// int Write_header
// role   : ecrit le hearder dans le fichier compresse
// entree : nombre de caractere differents dans la table
// sortie : OK si pas de probleme, DEST_ERROR si probleme d'ecriture
/*******************************************************/
int Write_header(unsigned int nb_car)
{
	S_Header header;		// creation d'un header


				// initialisation du header
	header.sign = 0x8F;											// tag
	strcpy(header.nom_original, source.nom);					// nom_original
	header.taille_originale =(unsigned long) source.length;		// taille originale
	header.taille_table =(unsigned int) nb_car;					// taille table


				// creation du fichier compresse
	if( (destination.handle = (int) open(destination.nom, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, S_IWRITE)) == -1)
	{
		return DEST_ERROR;
	}

				// ecriture du header
	if( write(destination.handle, &header, sizeof(header)) == -1)
	{
		close(destination.handle);
		return DEST_ERROR;
	}

	return OK;
}


/*******************************************************/
// int Write_table
// role   : ecrit la table de code dans le fichier compresse
// entree : rien
// sortie : OK si pas de probleme, DEST_ERROR si probleme d'ecriture
/*******************************************************/
int Write_table(void)
{

	unsigned char code = 0;				// code hexa du caractere
	unsigned char length_code = 0;		// nombre de bit de ce code
	unsigned char caractere = 0;
	unsigned char cpt2 = 0;
	unsigned int cpt = 0;


							// balaye toute la table de code
	for(cpt=0, caractere=0; cpt<256; cpt++, caractere++)
	{
		if( strcmp(table_code[cpt], "2") )	// si la cellule contient un code
		{
			if( write(destination.handle, &caractere, sizeof(unsigned char)) == -1)		// ecrit le caractere correspondant
				return DEST_ERROR;

			length_code = strlen(table_code[cpt]);					// recupere le nombre de bit du code
			if( write(destination.handle, &length_code, sizeof(unsigned char)) == -1)		// l'ecrit
				return DEST_ERROR;

			for(cpt2=0; cpt2<=(length_code>>3); cpt2++)			// transforme le code Ascii en hexa par tranche de 8 bits
			{
				code = Ascii2hex(table_code[cpt], cpt2);
				if( write(destination.handle, &code, sizeof(unsigned char)) == -1)		// ecrit le code hexa
					return DEST_ERROR;
			}
		}
	}
	return OK;
}


/*******************************************************/
// unsigned char Ascii2hex
// role   : transforme une chaine Ascii (n'ayant comme caractere que des 0 et 1) en nombre hexadecimal
// entree : chaine de caractere, a partir de quel bit on commence
// sortie : code hexadecimal
/*******************************************************/

unsigned char Ascii2hex(unsigned char code[], unsigned char offset)
{
	unsigned short i=0;					// compteur
	unsigned char code_hex=0;			// code final en hexa
	signed int length=0;				// taille du code ASCII

	length = strlen(code)-(offset<<3);	// recupere le nombre de bit du code
	
	for(i=0; (i<8) && (i<length); i++)	// on traite 8 bits au max
	{
		if (code[i+(offset<<3)]== '1')				// effectue un masquage binaire
			code_hex |= (1<<i);
	}

return code_hex;
}



/*******************************************************/
// int Write_file
// role   : ecrit le fichier compresse
// entree : rien
// sortie : OK si pas de probleme, DEST_ERROR si probleme d'ecriture
/*******************************************************/
int Write_file()
{

	unsigned long lcpt = 0L;					// pointeur long
	unsigned char car;
	unsigned char code_ascii[255+7+1];			// code Ascii pret a etre transforme
	unsigned char code_hexa;					// code hexa devant etre ecrit
	unsigned char scar_read[BUFFER_LENGTH +1];	// tampon de caractere lus
	unsigned char scar_write[BUFFER_LENGTH +1];	// tampon de code devant etre ecrit
	register signed int i=0, cpt=0, j=0;		// compteurs
	signed int nb_octet_lu = 0;

										// initialise les tampons
	memset(code_ascii,'\0', 263);
	memset(scar_write, '\0', BUFFER_LENGTH);

	lseek(source.handle, 0L, SEEK_SET);			// remet le pointeur du fichier source au debut de celui-ci

												// traite tout le fichier source
	for (lcpt=0L; lcpt<(long)source.length; lcpt+=(unsigned long) BUFFER_LENGTH)
	{
		if( (nb_octet_lu = read(source.handle, scar_read, BUFFER_LENGTH)) == -1)	// tente de lire BUFFER_LENGTH caracteres
			return SOURCE_ERROR;													// recupere le nombre de caracteres lus


		for(cpt=0; cpt<nb_octet_lu; cpt++)		// compresse tous les caracteres du tampon
		{
			strcat(code_ascii, table_code[scar_read[cpt]]);		// ajoute le ccde

			while (strlen(code_ascii) > 7)		// tant que le code fait plus de 8 bits
			{
				code_hexa = Ascii2hex(code_ascii);			// converti en hexa

				scar_write[j++] =(unsigned char) code_hexa;	// rajoute le code hexa dans le tampon pour etre ecrit
				if( j == BUFFER_LENGTH)		// si le tampon d'ecriture fait plus de BUFFER_LENGTH caracteres, on l'ecrit
				{
					if( write(destination.handle,scar_write,BUFFER_LENGTH) == -1)	// ecrit le tampon d'ecriture
						return DEST_ERROR;
					memset(scar_write, '\0', BUFFER_LENGTH+1);						// puis le vide
					j=0;
				}

											// supprime les 8 bits que l'on vient de convertir du code Ascii
				for (i=0; code_ascii[i] != '\0'; code_ascii[i]=code_ascii[i+8],i++);
			}
		}
		gotoxy(curpos.x, curpos.y);				// affiche l'avancement
		printf("  %u\%", 100*lcpt/source.length);
	}

										// ecrit le reste du code du dernier caractere
	while(strlen(code_ascii))
	{
		code_hexa = Ascii2hex(code_ascii);
		scar_write[j++] =(unsigned char) code_hexa;
						// supprime les 8 bits que l'on vient de convertir du code Ascii
		for (i=0; code_ascii[i] != '\0'; code_ascii[i]=code_ascii[i+8],i++);
	}

	if( write(destination.handle,scar_write, j) == -1)
		return DEST_ERROR;


	return OK;
}

	


